import json
import uuid
from pytest_sanic.utils import TestClient
from typing import Dict, Text, List

import pytest
from sanic.testing import SanicTestClient
from sqlalchemy.orm import Session

from rasax.community.database.conversation import Conversation
from rasax.community.services.event_service import EventService
from rasax.community.tracker_utils import ACTION_SESSION_START_NAME
from tests.unit.conftest import user_event, bot_event, action_event

CONVERSATIONS_URL = "/api/conversations"


@pytest.fixture()
def conversation(event_service: EventService) -> Text:
    sender_id = uuid.uuid4().hex

    events = [
        action_event(sender_id, 1000, action_name=ACTION_SESSION_START_NAME),
        bot_event(sender_id, 2000),
        user_event(sender_id, 3000),
        bot_event(sender_id, 4000),
        action_event(sender_id, 5000),
    ]

    for event in events:
        event_service.save_event(json.dumps(event), sender_id=sender_id)

    event_service.add_flagged_message(sender_id, 3000)
    event_service.add_flagged_message(sender_id, 5000)

    return sender_id


async def test_events_metadata_id(
    test_client: SanicTestClient,
    auth_header: Dict[Text, Text],
    session: Session,
    conversation: Text,
):
    response = await test_client.get(
        f"{CONVERSATIONS_URL}/{conversation}", headers=auth_header
    )
    response_body = await response.json()
    received_ids = [e["metadata"]["rasa_x_id"] for e in response_body["events"]]

    conversation_entity = session.query(Conversation).get(conversation)
    expected_ids = [e.id for e in conversation_entity.events]

    assert received_ids == expected_ids


async def test_events_metadata_is_flagged(
    test_client: SanicTestClient,
    auth_header: Dict[Text, Text],
    session: Session,
    conversation: Text,
):
    response = await test_client.get(
        f"{CONVERSATIONS_URL}/{conversation}", headers=auth_header
    )
    response_body = await response.json()
    received_flagged_ids = [
        e["metadata"]["rasa_x_id"]
        for e in response_body["events"]
        if e["metadata"]["rasa_x_flagged"]
    ]

    conversation_entity = session.query(Conversation).get(conversation)
    expected_flagged_ids = [e.id for e in conversation_entity.events if e.is_flagged]

    assert received_flagged_ids == expected_flagged_ids


@pytest.mark.parametrize(
    "exclude_leading_action_session_start,expected_timestamps",
    [("true", [2000, 3000, 4000, 5000]), ("false", [1000, 2000, 3000, 4000, 5000])],
)
async def test_events_with_exclude_action_session_start(
    test_client: SanicTestClient,
    auth_header: Dict[Text, Text],
    conversation: Text,
    exclude_leading_action_session_start: Text,
    expected_timestamps: List[float],
):
    response = await test_client.get(
        f"{CONVERSATIONS_URL}/{conversation}",
        headers=auth_header,
        params={
            "exclude_leading_action_session_start": exclude_leading_action_session_start
        },
    )
    response_body = await response.json()

    assert [
        event["timestamp"] for event in response_body["events"]
    ] == expected_timestamps


@pytest.mark.parametrize(
    "accept_header_value,output_contains,content_type",
    [
        ("", 'sender_id":"{conversation_id}"', "application/json"),
        ("application/json", 'sender_id":"{conversation_id}"', "application/json"),
        ("text/markdown", "* test\n    - action\n", "text/markdown"),
    ],
)
async def test_get_events_with_dumps(
    test_client: TestClient,
    auth_header: Dict[Text, Text],
    conversation: Text,
    accept_header_value: Text,
    output_contains: Text,
    content_type: Text,
):
    headers = auth_header.copy()
    headers.update({"Accept": accept_header_value})
    response = await test_client.get(
        f"{CONVERSATIONS_URL}/{conversation}", headers=headers,
    )

    # check status code
    assert response.status == 200

    # make sure content_type header is correct
    assert response.headers.get("Content-Type") == content_type

    # check content of request body
    expected_content = output_contains.format(conversation_id=conversation)
    output = await response.text()
    assert expected_content in output
