from typing import Text, Dict, Any, List
import hashlib
import json
import asyncio
import uuid
import socket
from unittest import mock

from rasax.community.metrics import (
    REPOSITORY_CREATED_EVENT,
    GIT_CHANGES_PUSHED_EVENT,
    FEATURE_FLAG_UPDATED_EVENT,
)
from sqlalchemy.orm import Session
import pytest
from _pytest.monkeypatch import MonkeyPatch
from aioresponses import aioresponses, CallbackResult
from rasa.nlu.training_data.formats import MarkdownReader
from rasax.community import config, metrics
from rasax.community.services.data_service import DataService
from rasax.community.services.domain_service import DomainService
from rasax.community.services.story_service import StoryService
from rasax.community.services.nlg_service import NlgService
from rasax.community.services.model_service import ModelService
from rasax.community.services.event_service import EventService
from rasax.community.services.config_service import ConfigService, ConfigKey
from rasax.community.services.settings_service import SettingsService
from rasax.community.constants import (
    CONFIG_METRICS_ENABLED,
    CONFIG_METRICS_USER,
    RASA_PRODUCTION_ENVIRONMENT,
)
from rasax.community.version import __version__
from tests.unit.conftest import insert_test_model, user_event, reset_config_service
from tests.unit.services.test_domain_service import TEST_DOMAIN
from tests.unit.services.test_data_service import EXAMPLE_DATA_MD
from tests.unit.services.test_story_service import TEST_STORY
from tests.unit.services.test_settings_service import default_environments_config

METRICS_TEST_USER = "metrics-test-user"


def _mock_track_internal_exception(*args, **kwargs) -> None:
    raise Exception("Tracking failed")


@pytest.fixture()
def clean_config_service(config_service: ConfigService) -> None:
    # Clean before testing
    reset_config_service(config_service)
    yield
    # Clean up after testing (in case other tests use the service without resetting it)
    reset_config_service(config_service)


def test_initialize_configuration_from_file_server_mode(monkeypatch: MonkeyPatch):
    monkeypatch.setattr(config, "LOCAL_MODE", False)

    metrics.initialize_configuration_from_file(None)
    assert config.metrics_collection_config == {}


def test_initialize_configuration_from_database_local_mode(monkeypatch: MonkeyPatch):
    monkeypatch.setattr(config, "LOCAL_MODE", True)

    metrics.initialize_configuration_from_db(None)
    assert config.metrics_collection_config == {}


def test_initialize_configuration_from_database_clean(
    clean_config_service: None,
    session: Session,
    config_service: ConfigService,
    monkeypatch: MonkeyPatch,
):
    # Clean config: no environment variables and default config values in DB
    # Telemetry should be enabled

    monkeypatch.setattr(config, "metrics_collection_config", {})
    monkeypatch.delenv(ConfigKey.TELEMETRY_ENABLED, raising=False)
    monkeypatch.delenv(ConfigKey.TELEMETRY_UUID, raising=False)

    metrics.initialize_configuration_from_db(session)

    assert metrics.are_metrics_enabled()
    assert metrics.metrics_id() == config_service.get_value(ConfigKey.TELEMETRY_UUID)


def test_initialize_configuration_from_database_disabled(
    clean_config_service: None,
    session: Session,
    config_service: ConfigService,
    monkeypatch: MonkeyPatch,
):
    # Telemetry has been disabled by setting the configuration option in the database
    # No env variables
    # Telemetry should be disabled

    config_service.set_value(ConfigKey.TELEMETRY_ENABLED, False)

    monkeypatch.setattr(config, "metrics_collection_config", {})
    monkeypatch.delenv(ConfigKey.TELEMETRY_ENABLED, raising=False)

    metrics.initialize_configuration_from_db(session)

    assert not metrics.are_metrics_enabled()


def test_initialize_configuration_from_database_env_enabled(
    clean_config_service: None,
    session: Session,
    config_service: ConfigService,
    monkeypatch: MonkeyPatch,
):
    # Telemetry has been enabled via environment (in most cases: upgrading
    # from 0.23 and using same execution context to start the server)
    # New Rasa X 0.24 installs should not configure telemetry via environment!
    # Telemetry should be enabled

    telemetry_id = uuid.uuid4().hex
    monkeypatch.setattr(config, "metrics_collection_config", {})
    monkeypatch.setenv(ConfigKey.TELEMETRY_ENABLED, "true")
    monkeypatch.setenv(ConfigKey.TELEMETRY_UUID, telemetry_id)

    metrics.initialize_configuration_from_db(session)

    assert metrics.are_metrics_enabled()
    assert metrics.metrics_id() == telemetry_id


@pytest.mark.parametrize("metrics_consent", ["", "false", "foobar"])
def test_initialize_configuration_from_database_env_disabled(
    metrics_consent: Text,
    clean_config_service: None,
    session: Session,
    config_service: ConfigService,
    monkeypatch: MonkeyPatch,
):
    # Same test as previous one, but telemetry has been explicitly disabled via
    # environment variables.
    # Telemetry should be disabled

    monkeypatch.setattr(config, "metrics_collection_config", {})
    monkeypatch.setenv(ConfigKey.TELEMETRY_ENABLED, metrics_consent)

    metrics.initialize_configuration_from_db(session)

    assert not metrics.are_metrics_enabled()


async def test_track_async(enable_metrics: None, monkeypatch: MonkeyPatch):
    calls = 0

    def _response(url, **kwargs):
        nonlocal calls
        calls += 1
        return CallbackResult(body=json.dumps({"success": True}))

    # If we're running on Travis CI, then no POST requests should be sent
    monkeypatch.setenv("TRAVIS", "true")
    monkeypatch.setenv("CI", "true")

    with aioresponses() as responses:
        responses.post(metrics.SEGMENT_ENDPOINT, callback=_response)
        await metrics._track_async(METRICS_TEST_USER, "Test", {"foo": "bar"})

    assert calls == 0  # No actual POST requests done

    # If not on CI, then a POST request should have been made
    monkeypatch.delenv("TRAVIS")
    monkeypatch.delenv("CI")

    with aioresponses() as responses:
        responses.post(metrics.SEGMENT_ENDPOINT, callback=_response)
        await metrics._track_async(METRICS_TEST_USER, "Test", {"foo": "bar"})

    assert calls == 1

    # If the request triggers an exception, the exception should be logged and
    # not propagated
    with aioresponses() as responses:
        responses.post(metrics.SEGMENT_ENDPOINT, exception=socket.gaierror())
        await metrics._track_async(METRICS_TEST_USER, "Test", {"foo": "bar"})
        # If no exception was raised, the test has passed


def test_track_internal_loop_stopped(enable_metrics: None, monkeypatch: MonkeyPatch):
    # Test that _track_internal works correctly when there's no current event
    # loop running.
    magic_mock = mock.MagicMock()

    @asyncio.coroutine
    def _mock_track_async(*args, **kwargs):
        magic_mock(*args, **kwargs)

    monkeypatch.setattr(metrics, "_track_async", _mock_track_async)
    loop = asyncio.get_event_loop()
    assert not loop.is_running()

    metrics.track("Test", {"foo": "bar"}, {"quux": "baz"})

    assert magic_mock.call_count == 1
    assert magic_mock.call_args[0] == (
        METRICS_TEST_USER,
        "Test",
        {"foo": "bar", "metrics_id": METRICS_TEST_USER},
        {"quux": "baz"},
    )


async def test_track_internal_loop_running(
    enable_metrics: None, monkeypatch: MonkeyPatch
):
    # Test that _track_internal works correctly when there's an event loop
    # already running.
    magic_mock = mock.MagicMock()

    @asyncio.coroutine
    def _mock_track_async(*args, **kwargs):
        magic_mock(*args, **kwargs)

    monkeypatch.setattr(metrics, "_track_async", _mock_track_async)
    loop = asyncio.get_event_loop()
    assert loop.is_running()

    metrics.track("Test", {"foo": "bar"}, {"quux": "baz"})

    # Since _track_internal() is fire-and-forget when a global event loop is
    # running, wait a second before running the assertions to make sure the
    # funcion has been executed (and therefore the mock used).
    await asyncio.sleep(1)

    assert magic_mock.call_count == 1
    assert magic_mock.call_args[0] == (
        METRICS_TEST_USER,
        "Test",
        {"foo": "bar", "metrics_id": METRICS_TEST_USER},
        {"quux": "baz"},
    )


def test_track_ignore_exception(enable_metrics: None, monkeypatch: MonkeyPatch):
    monkeypatch.setattr(metrics, "_track_internal", _mock_track_internal_exception)

    # If the test finishes without raising any exceptions, then it's successful
    metrics.track("Test")


def test_get_project_status_event_empty(
    session: Session,
    domain_service: DomainService,
    data_service: DataService,
    story_service: StoryService,
    new_project: Text,
):
    actual = metrics._get_project_status_event(session, new_project)

    assert actual == {
        "project": hashlib.sha256(new_project.encode("utf-8")).hexdigest(),
        "local_mode": False,
        "rasa_x": __version__,
        "num_intent_examples": 0,
        "num_entity_examples": 0,
        "num_actions": 0,
        "num_templates": 0,
        "num_slots": 0,
        "num_forms": 0,
        "num_intents": 0,
        "num_entities": 0,
        "num_stories": 0,
        "num_conversations": 0,
        "num_models": 0,
        "num_lookup_table_files": 0,
        "num_lookup_table_entries": 0,
        "num_synonyms": 0,
        "num_regexes": 0,
        "num_environments": 0,
        "environment_names": [],
        "num_live_environments": 0,
    }


async def test_get_project_status_event(
    session: Session,
    domain_service: DomainService,
    data_service: DataService,
    story_service: StoryService,
    nlg_service: NlgService,
    event_service: EventService,
    new_project: Text,
    user: Dict[Text, Any],
):
    domain_service.store_domain(TEST_DOMAIN, new_project)
    domain_id = domain_service.get_domain_id(new_project)
    assert domain_id is not None

    nlg_service.replace_templates(
        [{"template": "utter_greet", "text": "Hi there!"}], user["username"], domain_id
    )

    training_data = MarkdownReader().reads(EXAMPLE_DATA_MD)
    data_service.replace_additional_training_features(new_project, training_data)

    model_service = ModelService("path", session)
    insert_test_model(new_project, "metrics_test_model", model_service)

    await story_service.save_stories(TEST_STORY, user["team"])

    test_event = user_event("metrics_test_user")
    event_service.save_event(json.dumps(test_event))

    settings_service = SettingsService(session)
    settings_service.save_environments_config(
        new_project, default_environments_config()["environments"]
    )

    session.commit()

    actual = metrics._get_project_status_event(session, new_project)
    actual["environment_names"].sort()

    assert actual == {
        "project": hashlib.sha256(new_project.encode("utf-8")).hexdigest(),
        "local_mode": False,
        "rasa_x": __version__,
        "num_intent_examples": 0,
        "num_entity_examples": 0,
        "num_actions": 1,
        "num_templates": 1,
        "num_slots": 3,
        "num_forms": 1,
        "num_intents": 4,
        "num_entities": 2,
        "num_stories": 2,
        "num_conversations": 1,
        "num_models": 1,
        "num_lookup_table_files": 1,
        "num_lookup_table_entries": 15,
        "num_synonyms": 2,
        "num_regexes": 1,
        "num_environments": 3,
        "environment_names": ["development", RASA_PRODUCTION_ENVIRONMENT, "worker"],
        "num_live_environments": 0,
    }


@pytest.mark.parametrize(
    "referrer, origin",
    [
        ("http://localhost/interactive", "interactive"),
        ("http://localhost/stories", "stories"),
    ],
)
def test_track_story_created(metrics_events: List, referrer: Text, origin: Text):
    metrics.track_story_created(referrer)

    assert metrics_events[0] == {
        "event": "Story Created",
        "properties": {"story_created_from": origin},
        "context": None,
    }


@pytest.mark.parametrize(
    "username, input_channel",
    [
        ("me", "rasa"),
        ("test-user", "rasa"),
        ("3d1693ea-9c51-4410-9e31-444e8a35727d", "Tester"),
        ("3d1693ea-9c51-4410-9e31-444e8a35727d", None),
    ],
)
def test_track_message_received(
    metrics_events: List, username: Text, input_channel: Text
):
    metrics.track_message_received(username, input_channel)

    assert metrics_events[0] == {
        "event": "Message Received",
        "properties": {
            "username": hashlib.sha256(username.encode("utf-8")).hexdigest(),
            "channel": input_channel or "rasa",
        },
        "context": None,
    }


@pytest.mark.parametrize(
    "referrer, origin",
    [
        ("http://localhost/conversations", "conversations"),
        (
            "http://localhost/conversations/bcf7bb85-0506-42fb-89d4-5c1e310c4133/13.16",
            "conversations",
        ),
        ("http://localhost/data", "annotate_new_data"),
    ],
)
def test_track_message_annotated_from_referrer(
    metrics_events: List, referrer: Text, origin: Text
):
    metrics.track_message_annotated_from_referrer(referrer)

    assert metrics_events[0] == {
        "event": "Message Annotated",
        "properties": {"message_annotated_from": origin},
        "context": None,
    }


def test_track_message_annotated(metrics_events: List):
    metrics.track_message_annotated(metrics.MESSAGE_ANNOTATED_CONVERSATIONS)

    assert metrics_events[0] == {
        "event": "Message Annotated",
        "properties": {
            "message_annotated_from": metrics.MESSAGE_ANNOTATED_CONVERSATIONS
        },
        "context": None,
    }


def test_track_metrics_disabled(metrics_events: List):
    config.metrics_collection_config[CONFIG_METRICS_ENABLED] = False
    metrics.track("Foobar")

    assert len(metrics_events) == 0


def test_track_metrics_without_user(metrics_events: List):
    config.metrics_collection_config[CONFIG_METRICS_USER] = None
    metrics.track("Foobar")

    assert len(metrics_events) == 0


def test_track_repository_creation(metrics_events: List):
    test_branch_name = "test_track_repository_created"
    metrics.track_repository_creation(test_branch_name)

    assert metrics_events[0] == {
        "event": REPOSITORY_CREATED_EVENT,
        "properties": {"branch": test_branch_name},
        "context": None,
    }


def test_git_changes_pushed(metrics_events: List):
    test_branch_name = "test_git_changes_pushed"
    metrics.track_git_changes_pushed(test_branch_name)

    assert metrics_events[0] == {
        "event": GIT_CHANGES_PUSHED_EVENT,
        "properties": {"branch": test_branch_name},
        "context": None,
    }


@pytest.mark.parametrize("enabled", [True, False])
def test_track_feature_flag(metrics_events: List, enabled: bool):
    feature_name = "test_track_feature_flag"

    metrics.track_feature_flag(feature_name, enabled)

    assert metrics_events[0] == {
        "event": FEATURE_FLAG_UPDATED_EVENT,
        "properties": {"name": feature_name, "enabled": enabled},
        "context": None,
    }
